import os
from monocypher._monocypher import ffi


__all__ = (
    'random',
    'copy_context',
    'crypto_wipe',
    'crypto_verify16',
    'crypto_verify32',
    'crypto_verify64',
    'ensure',
)


def random(n):
    """
    Generates exactly `n` random bytes.
    This just calls :py:func:`os.urandom` and returns the result.

    :rtype: :py:class:`bytes`
    """
    return os.urandom(n)


def copy_context(ctx_ptr, type):
    """
    Return a copy of the struct at `ctx_ptr` of the given `type`.
    This can be used to copy structs, e.g. ``crypto_blake2b_ctx``.
    This is equivalent to the following in C:

    .. code:: c

       type dst_ptr = malloc(sizeof(*ctx_ptr));
       memcpy(
           (void *) dst_ptr,
           (void *) ctx_ptr,
           sizeof(*ctx_ptr)
       );

    Example::

        >>> from monocypher.bindings import crypto_blake2b_init
        >>> ctx = crypto_blake2b_init()
        >>> u = copy_context(ctx, 'crypto_blake2b_ctx *')
        <cdata 'crypto_blake2b_ctx *' owning ...>
        >>> u == ctx
        False

    :param ctx_ptr: CFFI pointer to some struct.
    :param type: Type of `ctx_ptr`.
    """
    dst_ptr  = ffi.new(type)
    dst_void = ffi.cast('uint8_t *', dst_ptr)
    ctx_void = ffi.cast('uint8_t *', ctx_ptr)
    ffi.memmove(
        dst_void,
        ctx_void,
        ffi.sizeof(ctx_ptr[0]),
    )
    return dst_ptr


def ensure(cond, exc, *args):
    if not cond:
        raise exc(*args)


def ensure_length(name, value, length):
    ensure(
        len(value) == length,
        TypeError,
        '{} must have length {}'.format(name, length),
    )


def ensure_bytes_with_length(name, value, length):
    ensure(
        isinstance(value, bytes) and len(value) == length,
        TypeError,
        '{} must be bytes with length {}'.format(name, length),
    )


def ensure_range(name, value, min, max=float('+inf')):
    ensure(
        isinstance(value, int) and (min <= value <= max),
        TypeError,
        '{name} must be an integer between {min} and {max}'.format(name=name, min=min, max=max),
    )


class Key:
    __slots__ = ()

    def encode(self):
        return bytes(self)

    def __hash__(self):
        return hash(bytes(self))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return crypto_verify32(bytes(self), bytes(other))


from monocypher.bindings.crypto_utils import crypto_wipe, crypto_verify16, crypto_verify32, crypto_verify64  # noqa: E402



from monocypher.utils import copy_context
from monocypher.bindings.crypto_hash import (
    crypto_blake2b,
    crypto_blake2b_init, crypto_blake2b_update, crypto_blake2b_final,
    BLAKE2B_KEY_MIN, BLAKE2B_KEY_MAX,
    BLAKE2B_HASH_MIN, BLAKE2B_HASH_MAX,
)


__all__ = ('Blake2bContext', 'blake2b')


# Direct Interface
blake2b = crypto_blake2b


class Blake2bContext:
    """
    Can be used to incrementally compute the `blake2b` hash of a
    long stream of bytes (e.g. a large file) without having to read
    all of it into memory.
    Parameters have the same meaning as :py:func:`.blake2b`.
    This class is compatible with :py:mod:`hashlib`'s hash objects;
    see :py:mod:`hashlib` for details.
    """

    KEY_MIN  = BLAKE2B_KEY_MIN    #: Minimum Blake2b key length
    KEY_MAX  = BLAKE2B_KEY_MAX    #: Maximum Blake2b key length
    HASH_MIN = BLAKE2B_HASH_MIN   #: Minimum Blake2b digest length
    HASH_MAX = BLAKE2B_HASH_MAX   #: Maximum Blake2b digest length

    name = 'blake2b'
    block_size = 128

    __slots__ = ('_ctx', '_hash_size')

    def __init__(self, data=b'', key=b'', hash_size=64):
        self._ctx = crypto_blake2b_init(key=key, hash_size=hash_size)
        self._hash_size = hash_size
        self.update(data)

    @property
    def digest_size(self):
        return self._hash_size

    def _copy_ctx(self):
        return copy_context(self._ctx, 'crypto_blake2b_ctx *')

    def copy(self):
        obj = self.__class__.__new__(self.__class__)
        obj._ctx = self._copy_ctx()
        obj._hash_size = self._hash_size
        return obj

    def update(self, data):
        crypto_blake2b_update(self._ctx, data)

    def digest(self):
        # crypto_blake2b_final wipes the context on call
        return crypto_blake2b_final(self._copy_ctx())

    def hexdigest(self):
        return self.digest().hex()



from monocypher.utils import ensure_range
from monocypher._monocypher import lib, ffi


BLAKE2B_KEY_MIN = 0
BLAKE2B_KEY_MAX = 64
BLAKE2B_HASH_MIN = 1
BLAKE2B_HASH_MAX = 64


def crypto_blake2b(msg, key=b'', hash_size=64):
    ensure_range('len(key)', len(key), BLAKE2B_KEY_MIN, BLAKE2B_KEY_MAX)
    ensure_range('hash_size', hash_size, BLAKE2B_HASH_MIN, BLAKE2B_HASH_MAX)

    hash = ffi.new('uint8_t[]', hash_size)
    msg  = ffi.from_buffer('uint8_t[]', msg)
    key  = ffi.from_buffer('uint8_t[]', key)

    lib.crypto_blake2b_general(
        hash, hash_size,
        key, len(key),
        msg, len(msg),
    )
    return bytes(hash)


def crypto_blake2b_init(key=b'', hash_size=64):
    ensure_range('len(key)', len(key), BLAKE2B_KEY_MIN, BLAKE2B_KEY_MAX)
    ensure_range('hash_size', hash_size, BLAKE2B_HASH_MIN, BLAKE2B_HASH_MAX)

    ctx = ffi.new('crypto_blake2b_ctx *')
    key = ffi.from_buffer('uint8_t[]', key)
    lib.crypto_blake2b_general_init(ctx, hash_size, key, len(key))
    return ctx


def crypto_blake2b_update(ctx, msg):
    msg = ffi.from_buffer('uint8_t[]', msg)
    lib.crypto_blake2b_update(ctx, msg, len(msg))


def crypto_blake2b_final(ctx):
    hash = ffi.new('uint8_t[]', ctx.hash_size)
    lib.crypto_blake2b_final(ctx, hash)
    return bytes(hash)



from monocypher.utils import ensure_length
from monocypher._monocypher import lib, ffi


def crypto_lock(key, nonce, msg, ad=b''):
    """
    :returns: (bytes(mac), bytes(ciphertext))
    """
    ensure_length('key', key, 32)
    ensure_length('nonce', nonce, 24)

    key   = ffi.from_buffer('uint8_t[32]', key)
    nonce = ffi.from_buffer('uint8_t[24]', nonce)
    mac   = ffi.new('uint8_t[16]')
    ct    = ffi.new('uint8_t[]', len(msg))
    msg   = ffi.from_buffer('uint8_t[]', msg)
    ad    = ffi.from_buffer('uint8_t[]', ad)

    lib.crypto_lock_aead(
        mac,
        ct,
        key,
        nonce,
        ad, len(ad),
        msg, len(msg),
    )
    return bytes(mac), bytes(ct)


def crypto_unlock(key, mac, nonce, ciphertext, ad=b''):
    """
    :returns: None or bytes(msg)
    """
    ensure_length('key', key, 32)
    ensure_length('mac', mac, 16)
    ensure_length('nonce', nonce, 24)

    key   = ffi.from_buffer('uint8_t[32]', key)
    mac   = ffi.from_buffer('uint8_t[16]', mac)
    nonce = ffi.from_buffer('uint8_t[24]', nonce)
    ct = ffi.from_buffer('uint8_t[]', ciphertext)
    ad = ffi.from_buffer('uint8_t[]', ad)
    pt = ffi.new('uint8_t[]', len(ciphertext))

    rv = lib.crypto_unlock_aead(
        pt,
        key,
        nonce,
        mac,
        ad, len(ad),
        ct, len(ct),
    )
    if rv != 0:
        return None
    return bytes(pt)



from monocypher.utils import ensure_bytes_with_length, ensure, Key, random
from monocypher.bindings.crypto_public import crypto_key_exchange, crypto_key_exchange_public_key
from monocypher.secret import SecretBox


__all__ = ('PublicKey', 'PrivateKey', 'Box')


class PublicKey(Key):
    """
    X25519 public key. This can be published.

    :param pk: The public key (:py:class:`bytes`).
    """

    KEY_SIZE = 32  #: Length of a public key in bytes.

    __slots__ = ('_pk',)

    def __init__(self, pk):
        ensure_bytes_with_length('pk', pk, self.KEY_SIZE)
        self._pk = pk

    def __bytes__(self):
        return self._pk


class PrivateKey(Key):
    """
    X25519 private key. This **must** be kept secret.

    :param sk: The private key (:py:class:`bytes`).
    """

    KEY_SIZE = 32  #: Length of a private key in bytes.

    __slots__ = ('_sk',)

    def __init__(self, sk):
        ensure_bytes_with_length('sk', sk, self.KEY_SIZE)
        self._sk = sk

    @classmethod
    def generate(cls):
        """
        Generates a random :class:`.PrivateKey` object.

        :rtype: :class:`.PrivateKey`
        """
        return cls(random(cls.KEY_SIZE))

    @property
    def public_key(self):
        """
        Returns the corresponding :class:`.PublicKey` object.

        :rtype: :class:`.PublicKey`
        """
        return PublicKey(crypto_key_exchange_public_key(self._sk))

    def __bytes__(self):
        return self._sk


class Box(SecretBox):
    """
    A subclass of :class:`~monocypher.secret.SecretBox` object with the
    encryption key being the shared key computed from the key exchange.
    The shared key is computed using X25519 and HChaCha20.
    For details see `Monocypher's documentation <https://monocypher.org/manual/key_exchange>`_.

    :param your_sk: Your private key (a :class:`.PrivateKey` object).
    :param their_pk: Their public key (a :class:`.PublicKey` object).
    """

    __slots__ = ()

    def __init__(self, your_sk, their_pk):
        ensure(isinstance(your_sk, PrivateKey), TypeError, 'your_sk should be a PrivateKey instance')
        ensure(isinstance(their_pk, PublicKey), TypeError, 'their_pk should be a PublicKey instance')
        super().__init__(crypto_key_exchange(
            your_sk.encode(),
            their_pk.encode(),
        ))

    @property
    def shared_key(self):
        """
        Returns the shared secret. This value is safe for use as the key
        for other symmetric ciphers.
        """
        return self._key


class SealedBox:
    """
    SealedBox enables you to send messages decryptable only by the receipient.
    Each time a message is encrypted, a new ephemeral keypair is generated;
    the ephemeral private key is used in key-exchange with the `receipient_key`
    and thrown away after encryption.

    :param receipient_key: A :py:class:`.PublicKey` or :py:class:`.PrivateKey`
                           object. If a :py:class:`.PrivateKey` is provided,
                           then the SealedBox is able to decrypt messages.
                           Otherwise, SealedBox can only encrypt messages.
    """

    __slots__ = ('_pk', '_sk')

    def __init__(self, receipient_key):
        if isinstance(receipient_key, PrivateKey):
            self._pk = receipient_key.public_key
            self._sk = receipient_key
        elif isinstance(receipient_key, PublicKey):
            self._pk = receipient_key
            self._sk = None
        else:
            raise TypeError('receipient_key should be a PublicKey or PrivateKey instance')

    def encrypt(self, msg):
        """
        Encrypt the given `msg`. This works using a similar construction
        as that from libsodium's `crypto_box_seal <https://libsodium.gitbook.io/doc/public-key_cryptography/sealed_boxes>`_,
        but using Monocypher's high level functions.

        :param msg: The message to encrypt (bytes).
        :rtype: :py:class:`bytes`
        """
        ephemeral_sk = PrivateKey.generate()
        ephemeral_pk = ephemeral_sk.public_key

        nonce = bytes(Box.NONCE_SIZE)
        ct = Box(ephemeral_sk, self._pk).encrypt(msg, nonce=nonce).ciphertext
        return ephemeral_pk.encode() + ct

    def decrypt(self, ciphertext):
        """
        Decrypt the given `ciphertext`. Returns the original message if
        decryption was successful, otherwise raises :py:class:`~monocypher.secret.CryptoError`.
        If the provided key was a :py:class:`.PublicKey`, raises a
        :py:class:`RuntimeError`.

        :param ciphertext: The ciphertext to decrypt (bytes-like object).
        :rtype: :py:class:`bytes`
        """
        ensure(self._sk is not None, RuntimeError, 'SecretBox cannot decrypt using a PublicKey')
        e_pk = ciphertext[:PublicKey.KEY_SIZE]  # Ephemeral PublicKey
        ct   = ciphertext[PublicKey.KEY_SIZE:]  # MAC + encrypted message
        box  = Box(self._sk, PublicKey(e_pk))
        return box.decrypt(
            ciphertext=ct,
            nonce=bytes(Box.NONCE_SIZE),
        )



import os
import glob
import cffi


cwd = os.path.abspath(os.path.dirname(__file__))
sources = os.path.join(cwd, 'monocypher-3.1.2-src/')

cdefs_file = os.path.join(cwd, 'monocypher_exposed.h')
include_dir = sources
sources     = glob.glob(os.path.join(sources, '**/*.c'), recursive=True)

ffi = cffi.FFI()
ffi.set_source(
    '_monocypher',
    '''
#include <stdlib.h>
#include <monocypher.h>
#include <optional/monocypher-ed25519.h>
    ''',
    sources=sources,
    include_dirs=[include_dir],
)
with open(cdefs_file, 'r') as fp:
    ffi.cdef(fp.read())


if __name__ == "__main__":
    ffi.compile(verbose=True)



from monocypher.utils import ensure_length
from monocypher._monocypher import lib, ffi


def crypto_ed25519_public_key(secret_key):
    ensure_length('secret_key', secret_key, 32)

    sk = ffi.from_buffer('uint8_t[32]', secret_key)
    pk = ffi.new('uint8_t[32]')

    lib.crypto_ed25519_public_key(pk, sk)
    return bytes(pk)


def crypto_ed25519_sign(secret_key, msg):
    ensure_length('secret_key', secret_key, 32)

    sk  = ffi.from_buffer('uint8_t[32]', secret_key)
    msg = ffi.from_buffer('uint8_t[]', msg)
    sig = ffi.new('uint8_t[64]')
    pk  = ffi.new('uint8_t[32]')

    lib.crypto_ed25519_public_key(pk, secret_key)
    lib.crypto_ed25519_sign(sig, sk, pk, msg, len(msg))
    lib.crypto_wipe(pk, 32)
    return bytes(sig)


def crypto_ed25519_check(sig, public_key, msg):
    ensure_length('sig', sig, 64)
    ensure_length('public_key', public_key, 32)

    sig = ffi.from_buffer('uint8_t[64]', sig)
    pk  = ffi.from_buffer('uint8_t[32]', public_key)
    msg = ffi.from_buffer('uint8_t[]', msg)

    rv = lib.crypto_ed25519_check(sig, pk, msg, len(msg))
    return rv == 0


def crypto_from_ed25519_private(eddsa):
    ensure_length('eddsa', eddsa, 32)

    eddsa  = ffi.from_buffer('uint8_t[32]', eddsa)
    x25519 = ffi.new('uint8_t[32]')
    lib.crypto_from_ed25519_private(x25519, eddsa)
    return bytes(x25519)


def crypto_from_ed25519_public(eddsa):
    ensure_length('eddsa', eddsa, 32)

    eddsa  = ffi.from_buffer('uint8_t[32]', eddsa)
    x25519 = ffi.new('uint8_t[32]')
    lib.crypto_from_eddsa_public(x25519, eddsa)
    return bytes(x25519)



from monocypher.utils import ensure_length
from monocypher._monocypher import lib, ffi


def crypto_verify16(a, b):
    ensure_length('a', a, 16)
    ensure_length('b', b, 16)

    return lib.crypto_verify16(
        ffi.from_buffer('uint8_t[16]', a),
        ffi.from_buffer('uint8_t[16]', b),
    ) == 0


def crypto_verify32(a, b):
    ensure_length('a', a, 32)
    ensure_length('b', b, 32)

    return lib.crypto_verify32(
        ffi.from_buffer('uint8_t[32]', a),
        ffi.from_buffer('uint8_t[32]', b),
    ) == 0


def crypto_verify64(a, b):
    ensure_length('a', a, 64)
    ensure_length('b', b, 64)

    return lib.crypto_verify64(
        ffi.from_buffer('uint8_t[64]', a),
        ffi.from_buffer('uint8_t[64]', b),
    ) == 0


def crypto_wipe(buf):
    buf = ffi.from_buffer('uint8_t[]', buf, require_writable=True)
    lib.crypto_wipe(buf, len(buf))



from monocypher.utils import ensure_length
from monocypher._monocypher import lib, ffi


def crypto_curve_to_hidden(your_pk, tweak):
    ensure_length('your_pk', your_pk, 32)

    curve = ffi.from_buffer('uint8_t[32]', your_pk)
    hidden = ffi.new('uint8_t[32]')

    if lib.crypto_curve_to_hidden(hidden, curve, tweak): # pragma: no cover
        return None # unsuitable for hiding
    return bytes(hidden)


def crypto_hidden_to_curve(your_hidden_pk):
    ensure_length('your_hidden_pk', your_hidden_pk, 32)

    hidden = ffi.from_buffer('uint8_t[32]', your_hidden_pk)
    pk = ffi.new('uint8_t[32]')

    lib.crypto_hidden_to_curve(pk, hidden)
    return bytes(pk)


def crypto_hidden_key_pair(your_secret_seed):
    ensure_length('your_secret_seed', your_secret_seed, 32)

    seed = ffi.from_buffer('uint8_t[32]', your_secret_seed)
    sk = ffi.new('uint8_t[32]')
    hidden_pk = ffi.new('uint8_t[32]')

    lib.crypto_hidden_key_pair(hidden_pk, sk, seed)
    return bytes(hidden_pk), bytes(sk)



from monocypher.utils import ensure_bytes_with_length, Key, random
from monocypher.bindings.crypto_sign import (
    crypto_check, crypto_sign, crypto_sign_public_key,
    crypto_from_eddsa_private, crypto_from_eddsa_public,
)
from monocypher.public import PublicKey, PrivateKey


__all__ = ('SignatureError', 'SignedMessage', 'VerifyKey', 'SigningKey')


class SignatureError(Exception):
    pass


class SignedMessage(bytes):
    """
    A subclass of :py:class:`bytes` representing a signed message.
    By default, the signature will be prepended to the message.
    """

    @classmethod
    def from_parts(cls, sig, msg):
        obj = cls(sig + msg)
        obj._sig = sig
        obj._msg = msg
        return obj

    @property
    def sig(self):
        """
        Returns the signature part.

        :rtype: :py:class:`bytes`
        """
        return self._sig

    @property
    def msg(self):
        """
        Returns the message part.

        :rtype: :py:class:`bytes`
        """
        return self._msg


class VerifyKey(Key):
    """
    EdDSA public key. This can be published.

    :param pk: The public key (:py:class:`bytes`),
               should be :py:obj:`.KEY_SIZE` bytes long.
    """

    KEY_SIZE = 32  #: Length of a public key.
    SIG_SIZE = 64  #: Length of a signature.

    __slots__ = ('_pk',)

    def __init__(self, pk):
        ensure_bytes_with_length('pk', pk, self.KEY_SIZE)
        self._pk = pk

    def verify(self, signed, sig=None):
        """
        Verify the given `signed` message. If `sig` is `None`, then the signature
        is assumed to be prepended to `signed`. Return the original message if the
        verification succeeds, otherwise :py:class:`.SignatureError` is raised.

        :param signed: A bytes-like object, or a :py:class:`.SignedMessage` object.
        :param sig: None, or a :py:class:`bytes` object with length :py:obj:`.SIG_SIZE`.
        """
        msg = signed
        if sig is None:
            if len(signed) < self.SIG_SIZE:
                raise SignatureError('corrupted message')
            sig = signed[:self.SIG_SIZE]
            msg = signed[self.SIG_SIZE:]

        if not crypto_check(sig=sig, public_key=self._pk, msg=msg):
            raise SignatureError('invalid signature')
        return msg

    def __bytes__(self):
        return self._pk

    def to_public_key(self):
        """
        Converts from a :py:class:`.VerifyKey` to a
        :py:class:`~monocypher.public.PublicKey` object.
        See notes about using the same key for both signing
        and key-exchange in :py:obj:`.SigningKey.to_private_key`.

        :rtype: :py:class:`~monocypher.public.PublicKey`
        """
        return PublicKey(crypto_from_eddsa_public(self._pk))


class SigningKey(Key):
    """
    EdDSA private key. This should be kept secret.

    :param sk: The secret key (:py:class:`bytes`),
               should be :py:obj:`.KEY_SIZE` bytes long.
    """

    KEY_SIZE = 32  #: Length of a secret key.
    SIG_SIZE = 64  #: Length of a signature.

    __slots__ = ('_sk',)

    def __init__(self, sk):
        ensure_bytes_with_length('sk', sk, self.KEY_SIZE)
        self._sk = sk

    @classmethod
    def generate(cls):
        """
        Generates a random :py:class:`.SigningKey` object.

        :rtype: :py:class:`.SigningKey`
        """
        return cls(random(cls.KEY_SIZE))

    def sign(self, msg):
        """
        Signs the given `msg`.

        :rtype: :py:class:`.SignedMessage`
        """
        sig = crypto_sign(secret_key=self._sk, msg=msg)
        return SignedMessage.from_parts(sig=sig, msg=msg)

    def __bytes__(self):
        return self._sk

    @property
    def verify_key(self):
        """
        Return the corresponding :py:class:`.VerifyKey` object.

        :rtype: :py:class:`.VerifyKey`
        """
        return VerifyKey(crypto_sign_public_key(self._sk))

    def to_private_key(self):
        """
        Converts from a :py:class:`.SigningKey` to a
        :py:class:`~monocypher.public.PrivateKey` object.
        The conversion is one-way and deterministic.
        Note that although the conversion is sound, you
        should not (without good reason) use the same private
        key for signing and key-exchange.

        :rtype: :py:class:`~monocypher.public.PrivateKey`
        """
        return PrivateKey(crypto_from_eddsa_private(self._sk))



from setuptools import setup


version = '0.0.3'


install_requires = ['cffi>=1.14']
setup_requirements = ['setuptools'] + install_requires
tests_require = ['pytest==6.2.4', 'hypothesis==6.14.1', 'pytest-cov==2.12.1']
docs_require  = ['Sphinx>=3.*']


setup(
    name='monocypher-py',
    version=version,

    description='Python binding to the Monocypher library',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/eugene-eeo/monocypher-py',
    license='CC0 1.0 Universal',

    author='Eeo Jun',
    author_email='141bytes@gmail.com',
    python_requires='>=3.6',
    setup_requires=setup_requirements,
    install_requires=install_requires,
    extras_require={
        'tests': tests_require,
        'docs':  docs_require,
    },
    tests_require=tests_require,

    package_dir={"": "src"},
    packages=[
        "monocypher",
        "monocypher.bindings",
    ],

    # CFFI
    ext_package="monocypher",
    cffi_modules=["src/build.py:ffi"],
    zip_safe=False,
)



from monocypher.utils import ensure_length
from monocypher._monocypher import lib, ffi


def crypto_key_exchange(your_secret_key, their_public_key):
    ensure_length('your_secret_key', your_secret_key, 32)
    ensure_length('their_public_key', their_public_key, 32)

    sk = ffi.from_buffer('uint8_t[32]', your_secret_key)
    pk = ffi.from_buffer('uint8_t[32]', their_public_key)
    shared = ffi.new('uint8_t[32]')

    lib.crypto_key_exchange(shared, sk, pk)
    return bytes(shared)


def crypto_key_exchange_public_key(your_secret_key):
    ensure_length('your_secret_key', your_secret_key, 32)

    sk = ffi.from_buffer('uint8_t[32]', your_secret_key)
    pk = ffi.new('uint8_t[32]')

    lib.crypto_key_exchange_public_key(pk, sk)
    return bytes(pk)


def crypto_x25519(your_secret_key, their_public_key):
    ensure_length('your_secret_key', your_secret_key, 32)
    ensure_length('their_public_key', their_public_key, 32)

    sk = ffi.from_buffer('uint8_t[32]', your_secret_key)
    pk = ffi.from_buffer('uint8_t[32]', their_public_key)
    shared = ffi.new('uint8_t[32]')

    lib.crypto_x25519(shared, sk, pk)
    return bytes(shared)


def crypto_x25519_public_key(your_secret_key):
    ensure_length('your_secret_key', your_secret_key, 32)

    sk = ffi.from_buffer('uint8_t[32]', your_secret_key)
    pk = ffi.new('uint8_t[32]')

    lib.crypto_x25519_public_key(pk, sk)
    return bytes(pk)



from .crypto_aead import crypto_lock, crypto_unlock
from .crypto_utils import crypto_verify16, crypto_verify32, crypto_verify64, crypto_wipe
from .crypto_hash import (
    crypto_blake2b,
    crypto_blake2b_init, crypto_blake2b_update, crypto_blake2b_final,
    BLAKE2B_HASH_MIN, BLAKE2B_HASH_MAX,
    BLAKE2B_KEY_MIN, BLAKE2B_KEY_MAX,
)
from .crypto_public import (
    crypto_key_exchange, crypto_key_exchange_public_key,
    crypto_x25519, crypto_x25519_public_key,
)
from .crypto_pwhash import crypto_argon2i
from .crypto_sign import (
    crypto_sign_public_key, crypto_sign, crypto_check,
    crypto_from_eddsa_private, crypto_from_eddsa_public,
)
from .crypto_ed25519 import (
    crypto_ed25519_public_key, crypto_ed25519_sign, crypto_ed25519_check,
    crypto_from_ed25519_private, crypto_from_ed25519_public,
)
from .crypto_hidden import (
    crypto_curve_to_hidden, crypto_hidden_to_curve,
    crypto_hidden_key_pair,
)

__all__ = (
    'crypto_lock', 'crypto_unlock',
    'crypto_verify16', 'crypto_verify32', 'crypto_verify64',
    'crypto_blake2b',
    'crypto_blake2b_init', 'crypto_blake2b_update', 'crypto_blake2b_final',
    'BLAKE2B_HASH_MIN', 'BLAKE2B_HASH_MAX',
    'BLAKE2B_KEY_MIN', 'BLAKE2B_KEY_MAX',
    'crypto_key_exchange', 'crypto_key_exchange_public_key',
    'crypto_x25519', 'crypto_x25519_public_key',
    'crypto_argon2i',
    'crypto_sign_public_key', 'crypto_sign', 'crypto_check',
    'crypto_from_eddsa_private', 'crypto_from_eddsa_public',
    'crypto_wipe',
)



from monocypher.utils import ensure_range
from monocypher._monocypher import lib, ffi


def crypto_argon2i(
    password,
    salt,
    hash_size=64,
    nb_blocks=100000,
    nb_iterations=3,
    key=b'',
    ad=b'',
):
    ensure_range('len(salt)', len(salt), min=8)
    ensure_range('hash_size', hash_size, min=4)
    ensure_range('nb_blocks', nb_blocks, min=8)
    ensure_range('nb_iterations', nb_iterations, min=1)

    work_area = lib.malloc(nb_blocks * 1024)
    if work_area == ffi.NULL:  # pragma: no cover
        raise RuntimeError('malloc() returned NULL')

    try:
        password = ffi.from_buffer('uint8_t[]', password)
        salt = ffi.from_buffer('uint8_t[]', salt)
        hash = ffi.new('uint8_t[]', hash_size)
        key  = ffi.from_buffer('uint8_t[]', key)
        ad   = ffi.from_buffer('uint8_t[]', ad)
        lib.crypto_argon2i_general(
            hash, hash_size,
            work_area, nb_blocks,
            nb_iterations,
            password, len(password),
            salt, len(salt),
            key, len(key),
            ad, len(ad),
        )
        return bytes(hash)
    finally:
        lib.free(work_area)
